import numpy as np
from keras.datasets import imdb
from keras.preprocessing import sequence

# settings
num_words = 10000
kv_len_origin = 5
q_len = 150

# load data
(x_train, y_train), (x_val, y_val) = imdb.load_data(num_words=num_words)
X = np.concatenate([x_train, x_val])
X = np.array([piece[1:] for piece in X])  # abandon <START> which is also a default token
y = np.concatenate([y_train, y_val])

word_to_id = imdb.get_word_index()
word_to_id = {k: (v + 3) for k, v in word_to_id.items()}
word_to_id['<START>'] = 1
word_to_id['<UNK>'] = 2
print('Data loaded.')

# load emb
emb_index = dict()
f = open('./data/glove.840B.300d.txt')
count = 0
for line in f:
    values = line.strip().split()
    if len(values) == 301:
        word = values[0]
        coefs = np.asarray(values[1:], dtype='float32')
        emb_index[word] = coefs
f.close()
print('Found %d word vectors of glove.' % len(emb_index))

emb = np.zeros((num_words, 300))
cnt = 0
for word, i in word_to_id.items():
    if i < num_words:
        if word in emb_index.keys():
            emb[i] = emb_index[word]
        else:
            cnt += 1
print('Null word embeddings: %d, %.4f.' % (cnt, cnt / num_words))

# preprocess
X_kv = sequence.pad_sequences(X, maxlen=kv_len_origin)
X_kv = np.hstack([X_kv, np.repeat([[word_to_id['a'], word_to_id['b'], word_to_id[
                 'c'], word_to_id['d']]], repeats=len(X_kv), axis=0)])
X_q = sequence.pad_sequences(X, maxlen=q_len)

# save data
np.savez('./data.npz', X_kv=X_kv, X_q=X_q, y=y,
         emb=emb, kv_len_origin=kv_len_origin, train_test_split=len(y_train))
print('Data saved.')
