import pickle
import numpy as np

import tensorflow as tf
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import random_ops

from keras import backend as K
from keras.layers import Layer


def get_selected_words(x_single, score, id_to_word, k):
    selected_words = {}  # {location: word_id}

    selected = np.argsort(score)[-k:]
    selected_k_hot = np.zeros(400)
    selected_k_hot[selected] = 1.0

    x_selected = (x_single * selected_k_hot).astype(int)
    return x_selected


def create_dataset_from_score(x, scores, k, name='L2X'):

    print('Creating dataset with selected sentences...')

    with open('data/id_to_word.pkl', 'rb') as f:
        id_to_word = pickle.load(f)
    new_data = []
    new_texts = []
    for i, x_single in enumerate(x):
        x_selected = get_selected_words(x_single,
                                        scores[i], id_to_word, k)

        new_data.append(x_selected)

    np.save('data/x_val-%s.npy' % name, np.array(new_data))


def calculate_acc(pred, y):
    return np.mean(np.argmax(pred, axis=1) == np.argmax(y, axis=1))


class Maskout(Layer):

    def __init__(self, keep_rate, noise_shape=None, seed=None, **kwargs):
        super(Maskout, self).__init__(**kwargs)
        self.keep_rate = min(1., max(0., keep_rate))
        self.noise_shape = noise_shape
        self.seed = seed

    def call(self, inputs):
        if self.noise_shape is None:
            noise_shape = K.shape(inputs)
        else:
            batch_size = K.shape(inputs)[0]
            noise_shape = tf.concat([[batch_size], self.noise_shape], axis=0)

        random_tensor = random_ops.random_uniform(
            noise_shape, dtype=tf.float32) + self.keep_rate
        binary_tensor = math_ops.floor(random_tensor)
        outputs = inputs * binary_tensor

        return outputs

    def get_config(self):
        config = {'keep_rate': self.keep_rate,
                  'noise_shape': self.noise_shape,
                  'seed': self.seed}
        base_config = super(Maskout, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape
