import os
import json
import numpy as np

from explain import create_original_model
from utils import calculate_acc


def validate():
    print('Loading dataset...')

    pred_val = np.load('data/pred_val.npy')
    print('Creating model...')
    model = create_original_model()

    weights_name = [i for i in os.listdir(
        './models') if i.startswith('original')][0]
    model.load_weights('./models/' + weights_name,
                       by_name=True)

    print_str = 'The validation accuracy with selected 25 pixels is'

    def validate_pred(method_name):
        print(method_name)

        try:
            x_val_selected = np.load('./data/x_val-%s.npy' % method_name)
            new_pred_val = model.predict(
                x_val_selected, verbose=0, batch_size=128)

            val_acc = calculate_acc(new_pred_val, pred_val)
            print('   ', print_str, val_acc)
            return val_acc
        except FileNotFoundError as e:
            print('   ', 'FileNotFoundError')
            return 0

    acc_l2x = validate_pred('L2X')
    acc_l2x_y = validate_pred('L2X_y')
    acc_pretrain_y = validate_pred('L2X_pretrain_y')

    result_str = {
        'acc_l2x': float(acc_l2x),
        'acc_l2x_y': float(acc_l2x_y),
        'acc_pretrain_y': float(acc_pretrain_y),
    }

    f = open('./result.jsonl', 'a')
    f.write(json.dumps(result_str) + '\n')
    f.close()


if __name__ == '__main__':
    validate()
